function [B, s, u, v] = mondriaan(A, NumProcessors, Imbalance, Permutation)
% mondriaan       Uses the Mondriaan algorithm to assign the nonzeros
%                 of a given sparse matrix A to NumProcessors processors
%                 such that the required communication for the matrix-
%                 vector multiplication u = A*v is as small as possible,
%         while ensuring that the number of nonzeros assigned to each
%         processor has a maximum imbalance given by Imbalance.
%         If desired, a permutation of A may also be specified in the
%         variable Permutation: 0 = no permutation,
%                               1 = move cut rows/columns to the front
%                               2 = move cut rows/columns to the middle
%                               3 = move cut rows/columns to the back.
%         
%         Return values:
%             B = A (possibly permuted) copy of the matrix A where the numerical
%                 value of each nonzero is replaced by the index of the
%                 processor to which this nonzero was assigned by the
%                 Mondriaan algorithm.
%             s = A statistics vector with information about the
%                 partitioning: duration, imbalance, max communication and
%                 communication volume.
%             v = Processor indices to which the components of the vector v
%                 are assigned.
%             u = Processor indices to which the components of the vector u
%                 are assigned.
%         
%         Usage:
%             [B, s] = mondriaan(A, NumProcessors, Imbalance, Permutation);
%             [B, s, u, v] = mondriaan(A, NumProcessors, Imbalance, Permutation);
%

if (nargin < 4)
    Permutation = 0;
end

[B, s, u, v] = MatlabMondriaan(A, NumProcessors, Imbalance, -1, Permutation);

end
